import re
from typing import Any, Dict, List
import markdownify
def _normalize_multiline_quote(text: str) -> str:
    lines = text.splitlines()
    normalized_lines = []
    quote_block = []

    for line in lines + [""]:
        if line.startswith(">"):quote_block.append(line[1:].strip())
        else:
            if quote_block:
                normalized_lines.append("$" + "\n".join(quote_block) + "$")
                quote_block = []
            normalized_lines.append(line)
    return "\n".join(normalized_lines).strip()

class Track_parsed:
    _PATT = re.compile(
    r"(?P<pre>```(?P<pre_c>[\s\S]*?)```)"
    r"|(?P<bold>\*\*(?P<bold_c>.*?)\*\*)"
    r"|(?P<mono>`(?P<mono_c>.*?)`)"
    r"|(?P<italic>__(?P<italic_c>.*?)__)"
    r"|(?P<underline>--(?P<underline_c>.*?)--)"
    r"|(?P<link>\[(?P<link_text>.*?)\]\((?P<link_url>\S+?)\))"
    r"|(?P<quote>\$(?P<quote_c>[\s\S]*?)\$)"
    r"|(?P<quote_md>^>(?P<quote_md_c>.*?)(?:\n|$))"
    r"|(?P<strike>~~(?P<strike_c>.*?)~~)"
    r"|(?P<spoiler>\|\|(?P<spoiler_c>.*?)\|\|)",
    flags=re.DOTALL,
)
    _TYPE_MAP = {
        "pre": "Pre",
        "bold": "Bold",
        "mono": "Mono",
        "italic": "Italic",
        "underline": "Underline",
        "strike": "Strike",
        "spoiler": "Spoiler",
        "quote": "Quote",
        "quote_md": "Quote",
        "link": "Link",
        "mention": "MentionText",
    }
    @staticmethod
    def _utf16_len_java_style(s: str) -> int:
        return len(s.encode("utf-16-be")) // 2
    @staticmethod
    def _html2md(src: str) -> str:
        src = re.sub(r'<i>(.*?)</i>', r'||\1||', src, flags=re.DOTALL)
        src = re.sub(r'<span class="spoiler">(.*?)</span>', r'||\1||', src, flags=re.DOTALL)
        src = markdownify.markdownify(html=src).strip()
        src = src.replace("@@SPOILER@@", "||")
        return src
    def transcribe(self, src: str, mode: str = "MARKDOWN") -> Dict[str, Any]:
        if mode and mode.upper() == "HTML":
            src = self._html2md(src)
        src = _normalize_multiline_quote(src)
        payload_parts: List[Dict[str, Any]] = []
        normalized_text = src
        byte_offset = 0
        char_offset = 0
        matches = list(self._PATT.finditer(src))
        for m in matches:
            whole = m.group(0)
            start, end = m.span()
            adj_from = self._utf16_len_java_style(src[:start]) - byte_offset
            adj_char_from = start - char_offset
            gname = m.lastgroup
            if not gname:continue
            if gname == "link":
                inner = m.group("link_text") or ""
                link_href = m.group("link_url")
                if link_href.startswith("u0"):gname = "mention"
            else:
                inner = m.group(f"{gname}_c") or ""
                link_href = None
            if gname in ["quote", "quote_md", "bold", "italic", "underline", "spoiler", "strike","mention"]:
                inner_metadata = self.transcribe(inner, mode="MARKDOWN")
                inner = inner_metadata["text"]
                if "metadata" in inner_metadata:
                    for part in inner_metadata["metadata"]["meta_data_parts"]:
                        part["from_index"] += adj_from
                        payload_parts.append(part)
            if inner == "":
                continue
            content_utf16_len = self._utf16_len_java_style(inner)
            part: Dict[str, Any] = {
                "type": self._TYPE_MAP.get(gname, "Unknown"),
                "from_index": adj_from,
                "length": content_utf16_len,
            }
            if gname == "mention":
                part["type"] = "MentionText"
                part["mention_text_user_id"] = link_href
            elif link_href:
                part["type"] = "Link"
                part["link_url"] = link_href
            payload_parts.append(part)
            normalized_text = (
                normalized_text[:adj_char_from] + inner + normalized_text[end - char_offset :]
            )
            byte_offset += self._utf16_len_java_style(whole) - content_utf16_len
            char_offset += (end - start) - len(inner)

        result: Dict[str, Any] = {"text": normalized_text.strip()}
        if payload_parts:
            result["metadata"] = {"meta_data_parts": payload_parts}

        return result


    def parse(self, text: str, parse_mode: str = "MARKDOWN") -> Dict[str, Any]:
        return self.transcribe(text, mode=parse_mode)


class APIRequestError(Exception):
    """Base class for all API request errors."""
    def __init__(self, status: str, dev_message: str = None):
        self.status = status
        self.dev_message = dev_message
        msg = f"{self.__class__.__name__}: {status}"
        if dev_message:
            msg += f" | Message : {dev_message}"
        super().__init__(msg)


class InvalidAccessError(APIRequestError):
    """Raised when access is invalid."""

class InvalidTokenError(Exception):
    pass

class InvalidInputError(APIRequestError):
    """Raised when input is invalid."""


class TooRequestError(APIRequestError):
    """Raised when too many requests are made."""


def raise_for_status(response: dict):
    status = response.get("status")
    dev_message = response.get("dev_message")

    if status == "INVALID_ACCESS":
        raise InvalidAccessError(status, dev_message)
    elif status == "INVALID_INPUT":
        raise InvalidInputError(status, dev_message)
    elif status == "TOO_REQUEST":
        raise TooRequestError(status, dev_message)
    elif status != "OK":
        raise APIRequestError(status, dev_message)


from json import dumps, loads
from tqdm import tqdm
from urllib3 import PoolManager, ProxyManager
from ..utils import Configs
from ..exceptions import *
from .helper import Helper

class Network:

    def __init__(self, methods:object) -> None:
        self.methods = methods
        self.sessionData = methods.sessionData
        self.crypto = methods.crypto
        self.http = ProxyManager(methods.proxy) if methods.proxy else PoolManager()

    def request(self, method:str, input:dict={}, tmpSession:bool=False, attempt:int = 0, maxAttempt:int=2):
        url:str = Helper.getApiServer()
        platform:str = self.methods.platform.lower()
        apiVersion:int = self.methods.apiVersion

        if platform in ["rubx", "rubikax"]:
            client:dict = Configs.clients["android"]
            client["package"] = "ir.rubx.bapp"
        
        elif platform in ["android"]:
            client:dict = Configs.clients["android"]

        else:
            client:dict = Configs.clients["web"]

        data = {
            "api_version": str(apiVersion),
            "tmp_session" if tmpSession else
            "auth": self.crypto.auth if tmpSession else
            self.crypto.changeAuthType(self.sessionData["auth"]) if apiVersion > 5 else
            self.sessionData["auth"],
            "data_enc": self.crypto.encrypt(
                dumps({
                    "method": method,
                    "input": input,
                    "client": client
                })
            )
        }

        headers:dict = {
            "Referer": "https://web.rubika.ir/",
            "Content-Type": "application/json; charset=utf-8"
        }

        if not tmpSession and apiVersion > 5:
            data["sign"] = self.crypto.makeSignFromData(data["data_enc"])

        while True:
            result = self.http.request(
                method="POST",
                url=url,
                headers=headers,
                body = dumps(data).encode(),
                timeout=self.methods.timeOut
            )

            try:
                result = loads(self.crypto.decrypt(loads(result.data.decode("UTF-8"))["data_enc"]))
            except:
                attempt += 1

                if attempt > maxAttempt:
                    raise
                
                continue

            if result["status"] == "OK":
                if tmpSession:
                    result["data"]["tmp_session"] = self.crypto.auth

                return result["data"]
            
            else:
                raise {
                    "INVALID_AUTH": InvalidAuth(),
                    "NOT_REGISTERED": NotRegistered(),
                    "INVALID_INPUT": InvalidInput(),
                    "TOO_REQUESTS": TooRequests()
                }[result["status_det"]]

    def upload(self, file:str, fileName:str=None, chunkSize:int=131072):
        from ..utils import Utils

        if isinstance(file, str):
            if Utils.checkLink(url=file):
                file:bytes = self.http.request(method="GET", url=file).data
                mime:str = Utils.getMimeFromByte(bytes=file)
                fileName = fileName or Utils.generateFileName(mime=mime)
            else:
                fileName = fileName or file
                mime = file.split(".")[-1]
                file = open(file, "rb").read()

        elif not isinstance(file, bytes):
            raise FileNotFoundError("Enter a valid path or url or bytes of file.")
        else:
            mime = Utils.getMimeFromByte(bytes=file)
            fileName = fileName or Utils.generateFileName(mime=mime)

        def send_chunk(data, maxAttempts=2):
            for attempt in range(maxAttempts):
                try:
                    response = self.http.request(
                        "POST",
                        url=requestSendFileData["upload_url"],
                        headers=header,
                        body=data
                    )
                    return loads(response.data.decode("UTF-8"))
                except Exception:
                    print(f"\nError uploading file! (Attempt {attempt + 1}/{maxAttempts})")
            
            print("\nFailed to upload the file!")

        requestSendFileData:dict = self.methods.requestSendFile(
            fileName = fileName,
            mime = mime,
            size = len(file)
        )

        header = {
            "auth": self.sessionData["auth"],
            "access-hash-send": requestSendFileData["access_hash_send"],
            "file-id": requestSendFileData["id"],
        }

        totalParts = (len(file) + chunkSize - 1) // chunkSize

        if self.methods.showProgressBar:
            processBar = tqdm(
                desc=f"Uploading {fileName}",
                total=len(file),
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        for partNumber in range(1, totalParts + 1):
            startIdx = (partNumber - 1) * chunkSize
            endIdx = min(startIdx + chunkSize, len(file))
            header["chunk-size"] = str(endIdx - startIdx)
            header["part-number"] = str(partNumber)
            header["total-part"] = str(totalParts)
            data = file[startIdx:endIdx]
            hashFileReceive = send_chunk(data)
            
            if self.methods.showProgressBar:
                processBar.update(len(data))

            if not hashFileReceive:
                return
            
            if partNumber == totalParts:

                if not hashFileReceive["data"]:
                    return
                
                requestSendFileData["file"] = file
                requestSendFileData["access_hash_rec"] = hashFileReceive["data"]["access_hash_rec"]
                requestSendFileData["file_name"] = fileName
                requestSendFileData["mime"] = mime
                requestSendFileData["size"] = len(file)
                return requestSendFileData
            
    def download(self, accessHashRec:str, fileId:str, dcId:str, size:int, fileName:str, chunkSize:int=262143, attempt:int=0, maxAttempts:int=2):
        headers:dict = {
            "auth": self.sessionData["auth"],
            "access-hash-rec": accessHashRec,
            "dc-id": dcId,
            "file-id": fileId,
            "Host": f"messenger{dcId}.iranlms.ir",
            "client-app-name": "Main",
            "client-app-version": "3.5.7",
            "client-package": "app.rbmain.a",
            "client-platform": "Android",
            "Connection": "Keep-Alive",
            "Content-Type": "application/json",
            "User-Agent": "okhttp/3.12.1"
        }


        response = self.http.request(
            "POST",
            url=f"https://messenger{dcId}.iranlms.ir/GetFile.ashx",
            headers=headers,
            preload_content=False
        )

        data:bytes = b""

        if self.methods.showProgressBar:
            processBar = tqdm(
                desc=f"Downloading {fileName}",
                total=size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
            )

        for downloadedData in response.stream(chunkSize):
            try:
                if downloadedData:
                    data += downloadedData
                    if self.methods.showProgressBar:
                        processBar.update(len(downloadedData))

                if len(data) >= size:
                    return data
            except Exception:
                if attempt <= maxAttempts:
                    attempt += 1
                    print(f"\nError downloading file! (Attempt {attempt}/{maxAttempts})")
                    continue

                raise TimeoutError("Failed to download the file!")


"""
🔹 Synchronous Example
```python
from rubka import Bot,Message

bot = Robot(token="YOUR_BOT_TOKEN")

@bot.on_message(commands=["start", "help"])
def handle_start(bot: Robot, message: Message):
    message.reply("👋 Hello! Welcome to the Rubka bot (sync example).")

bot.run()
```
🔹 Asynchronous Example
```python
from rubka import Robot, Message

bot = Robot(token="YOUR_BOT_TOKEN")

@bot.on_message(commands=["start", "help"])
async def handle_start(bot: Robot, message: Message):
    await message.reply("⚡ Hello! This is the async version of Rubka.")

bot.run()
```
Explanation

Uses rubka.asynco.Robot for asynchronous operation.

The handler handle_start is defined with async def.

await message.reply(...) is non-blocking: the bot can process other tasks while waiting for Rubika’s response.

asyncio.run(main()) starts the async event loop.

This approach is more powerful and recommended for larger bots or when you:

Need to call external APIs.

Handle multiple long-running tasks.

Want better performance and scalability.

👉 In short:

Sync = simple, step-by-step, blocking.

Async = scalable, concurrent, non-blocking.
"""
import requests
from typing import Optional
try:
    from importlib.metadata import version, PackageNotFoundError
    use_importlib_metadata = True
except ImportError:
    import pkg_resources
    use_importlib_metadata = False
def get_installed_version(package_name: str) -> Optional[str]:
    if use_importlib_metadata:
        try:return version(package_name)
        except PackageNotFoundError:return None
    else:
        try:return pkg_resources.get_distribution(package_name).version
        except pkg_resources.DistributionNotFound:return None
def get_latest_version(package_name: str) -> Optional[str]:
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        response = requests.get(url, timeout=5)
        response.raise_for_status()
        data = response.json()
        return data.get("info", {}).get("version")
    except Exception as e:
        #print(f"Error fetching latest version: {e}")
        return None

def check_rubka_version():
    package_name = "rubka"
    installed_version = get_installed_version(package_name)
    if installed_version is None:
        print(f"{package_name} is not installed.")
        return
    latest_version = get_latest_version(package_name)
    if latest_version is None:
        print("Failed to fetch latest version info.")
        return
    if installed_version != latest_version:
        print("⚠️ **Warning: Outdated Version Detected!** ⚠️")
        print("This version poses potential risks to the stability, security, and compatibility of your system.")
        print(f"\n- Installed version: {installed_version}")
        print(f"- Latest available version: {latest_version}")
        print("\nImmediate action is required to ensure optimal performance and security.")
        print(f"To update, run the following command:")
        print(f"\n    pip install {package_name}=={latest_version}\n")
        print("For more details and updates, Channel : @rubka_info")
check_rubka_version()
from .asynco import Robot,Message,ChatKeypadBuilder,InlineBuilder,filters,InlineMessage
from .api import Robot as Bot,Message,InlineMessage
from .exceptions import APIRequestError
from .rubino import Bot as rubino
from .tv import TV as TvRubika

__all__ = [
    "Robot",
    "on_message",
    "APIRequestError",
    "create_simple_keyboard",
]



from typing import Callable,Union
import re
class TextFilter:
    def __call__(self, keyword=None):
        if keyword is None:
            return Filter(lambda m: getattr(m, "is_text", False))
        else:
            return Filter(lambda m: getattr(m, "is_text", False) and keyword in getattr(m, "text", ""))
class Filter:
    def __init__(self, func: Callable):
        self.func = func

    def __call__(self, message):
        return self.func(message)

    def __and__(self, other):
        return Filter(lambda m: self(m) and other(m))

    def __or__(self, other):
        return Filter(lambda m: self(m) or other(m))

    def __invert__(self):
        return Filter(lambda m: not self(m))

    def __xor__(self, other):
        return Filter(lambda m: self(m) != other(m))

    def __eq__(self, other):
        return Filter(lambda m: self(m) == other)

    def __ne__(self, other):
        return Filter(lambda m: self(m) != other)

    def __lt__(self, other):
        return Filter(lambda m: self(m) < other)

    def __le__(self, other):
        return Filter(lambda m: self(m) <= other)

    def __gt__(self, other):
        return Filter(lambda m: self(m) > other)

    def __ge__(self, other):
        return Filter(lambda m: self(m) >= other)

    def __add__(self, other):
        return Filter(lambda m: self(m) + (other(m) if callable(other) else other))

    def __sub__(self, other):
        return Filter(lambda m: self(m) - (other(m) if callable(other) else other))

    def __mul__(self, other):
        return Filter(lambda m: self(m) * (other(m) if callable(other) else other))

    def __truediv__(self, other):
        return Filter(lambda m: self(m) / (other(m) if callable(other) else other))
class IsCommand(Filter):
    def __init__(self, commands=None):
        if commands is None:
            func = lambda m: getattr(m, "is_command", False)
        else:
            if isinstance(commands, str):
                commands = [commands]
            func = lambda m: getattr(m, "is_command", False) and getattr(m, "text", "").lstrip("/").split()[0] in commands

        super().__init__(func)

    def __getattr__(self, name: str):
        return IsCommand([name])


class IsText:
    def __call__(self, text=None):
        if text is None:
            func = lambda m: m.is_text is True and m.is_file is False
        else:
            if isinstance(text, str):
                text = [text]
            func = lambda m: m.is_text is True and m.is_file is False and m.text in text

        return Filter(func)
#is_text = Filter(lambda m: getattr(m, "is_text", False))
is_file = Filter(lambda m: getattr(m, "file", None) is not None)
is_sticker = Filter(lambda m: getattr(m, "sticker", None) is not None)
is_contact = Filter(lambda m: getattr(m, "contact_message", None) is not None)
is_poll = Filter(lambda m: getattr(m, "poll", None) is not None)
is_location = Filter(lambda m: getattr(m, "location", None) is not None)
is_live_location = Filter(lambda m: getattr(m, "live_location", None) is not None)
has_any_media = Filter(lambda m: getattr(m, "has_any_media", False))
has_media = Filter(lambda m: getattr(m, "has_media", False))
is_text = IsText()
is_command = IsCommand()
is_user = Filter(lambda m: getattr(m, "is_user", False))
is_private = Filter(lambda m: getattr(m, "is_private", False))
is_group = Filter(lambda m: getattr(m, "is_group", False))
is_channel = Filter(lambda m: getattr(m, "is_channel", False))
is_reply = Filter(lambda m: getattr(m, "is_reply", False))
is_forwarded = Filter(lambda m: getattr(m, "is_forwarded", False))
is_edited = Filter(lambda m: getattr(m, "is_edited", False))
def text(keyword: str):
    return Filter(lambda m: getattr(m, "text", "") and keyword in m.text)
def text_length(min_len: int = 0, max_len: int = None):
    def _filter(m):
        t = getattr(m, "text", "")
        if not t: return False
        if len(t) < min_len: return False
        if max_len is not None and len(t) > max_len: return False
        return True
    return Filter(_filter)
def text_regex(pattern: str):
    regex = re.compile(pattern)
    return Filter(lambda m: getattr(m, "text", "") and regex.search(m.text))
def regex(pattern: str):
    regex = re.compile(pattern)
    return Filter(lambda m: getattr(m, "text", "") and regex.search(m.text))
def text_startswith(prefix: str):
    return Filter(lambda m: getattr(m, "text", "").startswith(prefix) if getattr(m, "text", None) else False)
def text_endswith(suffix: str):
    return Filter(lambda m: getattr(m, "text", "").endswith(suffix) if getattr(m, "text", None) else False)
def text_upper():
    return Filter(lambda m: getattr(m, "text", "").isupper() if getattr(m, "text", None) else False)
def text_lower():
    return Filter(lambda m: getattr(m, "text", "").islower() if getattr(m, "text", None) else False)
def text_digit():
    return Filter(lambda m: getattr(m, "text", "").isdigit() if getattr(m, "text", None) else False)
def text_word_count(min_words: int = 1, max_words: int = None):
    def _filter(m):
        t = getattr(m, "text", "")
        if not t: return False
        wc = len(t.split())
        if wc < min_words: return False
        if max_words is not None and wc > max_words: return False
        return True
    return Filter(_filter)
def text_contains_any(keywords: list):
    return Filter(lambda m: getattr(m, "text", "") and any(k in m.text for k in keywords))
def text_equals(value: str):
    return Filter(lambda m: getattr(m, "text", None) == value)
def text_not_equals(value: str):
    return Filter(lambda m: getattr(m, "text", None) != value)
def file_size_gt(size: int):
    return Filter(lambda m: m.file and getattr(m.file, "size", 0) > size)
def file_size_lt(size: int):
    return Filter(lambda m: m.file and getattr(m.file, "size", 0) < size)
def file_name_contains(substring: str):
    return Filter(lambda m: m.file and substring in getattr(m.file, "file_name", ""))
def file_extension(ext: str):
    return Filter(lambda m: m.file and getattr(m.file, "file_name", "").endswith(ext))
def file_id_is(file_id: str):
    return Filter(lambda m: m.file and getattr(m.file, "file_id", None) == file_id)
def sticker_id_is(sid: str):
    return Filter(lambda m: m.sticker and getattr(m.sticker, "sticker_id", None) == sid)
def sticker_emoji_is(emoji: str):
    return Filter(lambda m: m.sticker and getattr(m.sticker, "emoji", None) == emoji)
is_bold = Filter(lambda m: getattr(m, "is_bold", False))
is_italic = Filter(lambda m: getattr(m, "is_italic", False))
is_strike = Filter(lambda m: getattr(m, "is_strike", False))
is_underline = Filter(lambda m: getattr(m, "is_underline", False))
is_quote = Filter(lambda m: getattr(m, "is_quote", False))
is_spoiler = Filter(lambda m: getattr(m, "is_spoiler", False))
is_pre = Filter(lambda m: getattr(m, "is_pre", False))
is_mono = Filter(lambda m: getattr(m, "is_mono", False))
is_link_meta = Filter(lambda m: getattr(m, "is_link_meta", False))
has_metadata = Filter(lambda m: getattr(m, "has_metadata", False))
meta_links_contain = lambda keyword: Filter(lambda m: any(keyword in link for link in getattr(m, "meta_links", [])))
meta_link_positions_contain = lambda keyword: Filter(lambda m: any(keyword in link.get("url", "") for link in getattr(m, "meta_link_positions", [])))
meta_types_include = lambda types: Filter(lambda m: any(t in getattr(m, "meta_types", []) for t in types))
def poll_question_contains(keyword: str):
    return Filter(lambda m: m.poll and keyword in getattr(m.poll, "question", ""))
def poll_option_count(min_options: int = 1, max_options: int = None):
    def _filter(m):
        if not getattr(m, "poll", None): return False
        options = getattr(m.poll, "options", [])
        if len(options) < min_options: return False
        if max_options is not None and len(options) > max_options: return False
        return True
    return Filter(_filter)
def location_within(lat_min, lat_max, long_min, long_max):
    def _filter(m):
        loc = getattr(m, "location", None)
        if not loc: return False
        return lat_min <= getattr(loc, "lat", 0) <= lat_max and long_min <= getattr(loc, "long", 0) <= long_max
    return Filter(_filter)
def live_location_within(lat_min, lat_max, long_min, long_max):
    def _filter(m):
        loc = getattr(m, "live_location", None)
        if not loc: return False
        return lat_min <= getattr(loc, "lat", 0) <= lat_max and long_min <= getattr(loc, "long", 0) <= long_max
    return Filter(_filter)
def has_media_types(types: list):
    return Filter(lambda m: any(getattr(m, t, None) for t in types))

def message_id_is(mid: str):
    return Filter(lambda m: getattr(m, "message_id", None) == mid)

def is_reply_to_user(user_id: str):
    return Filter(lambda m: getattr(m, "reply_to_message_id", None) == user_id)

def is_forwarded_from(user_id: str):
    return Filter(lambda m: getattr(m.forwarded_from, "sender_id", None) == user_id if getattr(m, "forwarded_from", None) else False)

def edited_text_contains(keyword: str):
    return Filter(lambda m: getattr(m, "edited_text", "") and keyword in m.edited_text)

def aux_data_contains(key: str, value):
    return Filter(lambda m: getattr(m.aux_data, key, None) == value if getattr(m, "aux_data", None) else False)




def file_attr(attr_name):
    return Filter(lambda m: m.file and getattr(m.file, attr_name, None))

def sticker_attr(attr_name):
    return Filter(lambda m: m.sticker and getattr(m.sticker, attr_name, None))

def poll_attr(attr_name):
    return Filter(lambda m: m.poll and getattr(m.poll, attr_name, None))

def location_attr(attr_name):
    return Filter(lambda m: m.location and getattr(m.location, attr_name, None))

def live_location_attr(attr_name):
    return Filter(lambda m: m.live_location and getattr(m.live_location, attr_name, None))




file_size = file_attr("size")
file_name = file_attr("file_name")
sticker_id = sticker_attr("sticker_id")
poll_question = poll_attr("question")
location_lat = location_attr("lat")
location_long = location_attr("long")
live_location_lat = live_location_attr("lat")
live_location_long = live_location_attr("long")

_custom_filters = {}
def chat_title_contains(keyword: str):
    
    return Filter(lambda m: getattr(m, "chat", None) and keyword in getattr(m.chat, "title", ""))

def chat_title_equals(value: str):
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "title", "") == value)

def chat_id_is(sender_id: str):
    return Filter(lambda m: getattr(m, "chat_id", None) == sender_id)
def sender_id_is(sender_id: str):
    return Filter(lambda m: getattr(m, "sender_id", None) == sender_id)
def senders_id(sender_ids: Union[str, list]):
    if isinstance(sender_ids, list):
        return Filter(lambda m: getattr(m, "sender_id", None) in sender_ids)
    return Filter(lambda m: getattr(m, "sender_id", None) == sender_ids)
def chat_ids(sender_ids: Union[str, list]):
    if isinstance(sender_ids, list):
        return Filter(lambda m: getattr(m, "chat_id", None) in sender_ids)
    return Filter(lambda m: getattr(m, "chat_id", None) == sender_ids)
def text_contains(substring: str, case_sensitive: bool = False):
    def safe_text(m):
        text = getattr(m, "text", None)
        return text if isinstance(text, str) else ""
    if case_sensitive:return Filter(lambda m: substring in safe_text(m))
    else:return Filter(lambda m: substring.lower() in safe_text(m).lower())

def chat_member_count(min_count: int = 0, max_count: int = None):
    
    def _filter(m):
        c = getattr(m, "chat", None)
        if not c: return False
        count = getattr(c, "member_count", 0)
        if count < min_count: return False
        if max_count is not None and count > max_count: return False
        return True
    return Filter(_filter)

def chat_type_is(chat_type: str):
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "type", None) == chat_type)

def chat_username_contains(keyword: str):
    
    return Filter(lambda m: getattr(m, "chat", None) and keyword in getattr(m.chat, "username", ""))

def chat_username_equals(value: str):
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "username", "") == value)

def chat_has_link():
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "invite_link", None) is not None)

def chat_is_private():
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "type", None) in ["group", "channel"])

def chat_member_count_gt(count: int):
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "member_count", 0) > count)

def chat_member_count_lt(count: int):
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "member_count", 0) < count)

def chat_has_username():
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "username", None) is not None)

def chat_type_in(types: list):
    
    return Filter(lambda m: getattr(m, "chat", None) and getattr(m.chat, "type", None) in types)


def chat_title_regex(pattern: str):
    regex = re.compile(pattern)
    return Filter(lambda m: getattr(m, "chat", None) and regex.search(getattr(m.chat, "title", "")))

def chat_username_regex(pattern: str):
    regex = re.compile(pattern)
    return Filter(lambda m: getattr(m, "chat", None) and regex.search(getattr(m.chat, "username", "")))
def custom(name):
    def wrapper(func):
        _custom_filters[name] = Filter(func)
        return _custom_filters[name]
    return wrapper

def get_custom(name):
    return _custom_filters.get(name)

def and_(*filters):
    return Filter(lambda m: all(f(m) for f in filters))

def or_(*filters):
    return Filter(lambda m: any(f(m) for f in filters))

def not_(filter_):
    return Filter(lambda m: not filter_(m))


import sys
import subprocess

def install_and_import(package_name):
    try:
        __import__(package_name)
    except ModuleNotFoundError:
        print(f"Module '{package_name}' not found. Installing...")
        subprocess.check_call([sys.executable, "-m", "pip", "install", package_name])
    finally:
        globals()[package_name] = __import__(package_name)

install_and_import("websocket")

from websocket import WebSocketApp

from .helper import Helper
from json import dumps, loads
from threading import Thread
from ..types import Message
from ..exceptions import NotRegistered, TooRequests
from ..utils import Utils
from re import match
from time import sleep

class Socket:
    def __init__(self, methods) -> None:
        self.methods = methods
        self.handlers = {}

    ...



import requests
from typing import List, Optional, Dict, Any, Literal
from .exceptions import APIRequestError
from .adaptorrubka import Client as Client_get
from .logger import logger
from . import filters
from . import helpers
from typing import Callable
from .context import Message,InlineMessage
from typing import Optional, Union, Literal, Dict, Any
from pathlib import Path
import time
import datetime
import tempfile
from tqdm import tqdm
import os
API_URL = "https://botapi.rubika.ir/v3"
import mimetypes
import re
import sys
import subprocess
class InvalidTokenError(Exception):pass
def install_package(package_name):
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install", package_name], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        return True
    except Exception:return False

def get_importlib_metadata():
    try:
        from importlib.metadata import version, PackageNotFoundError
        return version, PackageNotFoundError
    except ImportError:
        if install_package("importlib-metadata"):
            try:
                from importlib_metadata import version, PackageNotFoundError
                return version, PackageNotFoundError
            except ImportError:
                return None, None
        return None, None

version, PackageNotFoundError = get_importlib_metadata()
def get_installed_version(package_name: str) -> str:
    if version is None:return "unknown"
    try:
        return version(package_name)
    except PackageNotFoundError:
        return None
def get_latest_version(package_name: str) -> str:
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        resp = requests.get(url, timeout=5)
        resp.raise_for_status()
        data = resp.json()
        return data["info"]["version"]
    except Exception:return None


def show_last_six_words(text):
    text = text.strip()
    return text[-6:]
import requests
from pathlib import Path
from typing import Union, Optional, Dict, Any, Literal
import tempfile
import os
import requests
from typing import Any, Dict, List, Optional


class Robot:
    """
    Main class to interact with Rubika Bot API.
    Initialized with bot token.
    """

    def __init__(self, token: str, session_name: Optional[str] = None, auth: Optional[str] = None, Key: Optional[str] = None,
                 platform: str = "web", web_hook: Optional[str] = None, timeout: int = 10, show_progress: bool = False):
        """
        Initializes the Rubika bot with the provided token and configuration.

        Parameters:
            token (str): Bot token for authentication with the Rubika Bot API.
            session_name (str, optional): Name of the session for client identification. Defaults to None.
            auth (str, optional): Additional authentication value for custom client connection. Defaults to None.
            Key (str, optional): Encryption or verification key if required. Defaults to None.
            platform (str, optional): The platform for execution ("web", "android", etc.). Defaults to "web".
            web_hook (str, optional): URL for the webhook. Defaults to None.
            timeout (int, optional): Timeout duration for HTTP requests in seconds. Defaults to 10.
            show_progress (bool, optional): Flag to show progress. Defaults to False.

        Example:
            >>> bot = Robot(token="BOT_TOKEN", platform="android", timeout=15)
        """
        self.token = token
        self.timeout = timeout
        self.auth = auth
        self.show_progress = show_progress
        self.session_name = session_name
        self.Key = Key
        self.platform = platform
        self.web_hook = web_hook
        self.hook = web_hook
        self._offset_id = None
        self.session = requests.Session()
        self.sessions: Dict[str, Dict[str, Any]] = {}
        self._callback_handlers: List[dict] = []
        self._message_handlers: List[dict] = []
        self._inline_query_handlers: List[dict] = []
        
        # Initialize the token
        self.geteToken()

        # Handle webhook setup
        if web_hook:
            self._set_webhook(web_hook)

    def _set_webhook(self, web_hook: str):
        """
        Sets the webhook URL and updates the bot endpoints.

        Parameters:
            web_hook (str): The URL of the webhook to be set.
        """
        try:
            json_url = requests.get(web_hook, timeout=self.timeout).json().get('url', web_hook)
            for endpoint_type in [
                "ReceiveUpdate",
                "ReceiveInlineMessage",
                "ReceiveQuery",
                "GetSelectionItem",
                "SearchSelectionItems"
            ]:
                print(self.update_bot_endpoint(self.web_hook, endpoint_type))
            self.web_hook = json_url
        except Exception as e:
            logger.error(f"Failed to set webhook from {web_hook}: {e}")
    
    def geteToken(self):
        """
        Retrieves the token for the bot.
        Implement the method as required.
        """
        pass

    def update_bot_endpoint(self, webhook: str, endpoint_type: str) -> str:
        """
        Updates the bot's endpoint.

        Parameters:
            webhook (str): The webhook URL.
            endpoint_type (str): The type of endpoint to update.

        Returns:
            str: The updated endpoint URL.
        """
        # Placeholder for actual endpoint update logic
        return f"{webhook}/{endpoint_type}"


    def _post(self, method: str, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Sends a POST request to the Rubika Bot API.

        Parameters:
            method (str): The API method to call.
            data (dict): The data to send in the request.

        Returns:
            dict: The response from the API.

        Raises:
            APIRequestError: If the API request fails or the response is invalid.
        """
        url = f"{API_URL}/{self.token}/{method}"
        try:
            response = self.session.post(url, json=data, timeout=self.timeout)
            response.raise_for_status()
            try:
                json_resp = response.json()
            except ValueError:
                logger.error(f"Invalid JSON response from {method}: {response.text}")
                raise APIRequestError(f"Invalid JSON response: {response.text}")
            
            if method != "getUpdates":
                logger.debug(f"API Response from {method}: {json_resp}")
                
            return json_resp
        except requests.RequestException as e:
            logger.error(f"API request failed for method {method}: {e}")
            raise APIRequestError(f"API request failed for {method}: {e}") from e

    def get_me(self) -> Dict[str, Any]:
        """Get info about the bot itself."""
        return self._post("getMe", {})

    def geteToken(self):
        """Check if the bot token is valid by calling the `getMe` method."""
        response = self.get_me()
        if response.get('status') != "OK":
            raise InvalidTokenError("The provided bot token is invalid or expired.")
    def on_message_private(
        self,
        chat_id: Optional[Union[str, List[str]]] = None,
        commands: Optional[List[str]] = None,
        filters: Optional[Callable[[Message], bool]] = None,
        sender_id: Optional[Union[str, List[str]]] = None,
        sender_type: Optional[str] = None,  
        allow_forwarded: bool = True,
        allow_files: bool = True,
        allow_stickers: bool = True,
        allow_polls: bool = True,
        allow_contacts: bool = True,
        allow_locations: bool = True,
        min_text_length: Optional[int] = None,
        max_text_length: Optional[int] = None,
        contains: Optional[str] = None,
        startswith: Optional[str] = None,
        endswith: Optional[str] = None,
        case_sensitive: bool = False
    ):
        """
        Sync decorator for handling only private messages with extended filters.
        """

        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                
                if not message.is_private:
                    return

                
                if chat_id:
                    if isinstance(chat_id, str) and message.chat_id != chat_id:
                        return
                    if isinstance(chat_id, list) and message.chat_id not in chat_id:
                        return

                
                if sender_id:
                    if isinstance(sender_id, str) and message.sender_id != sender_id:
                        return
                    if isinstance(sender_id, list) and message.sender_id not in sender_id:
                        return

                
                if sender_type and message.sender_type != sender_type:
                    return

                
                if not allow_forwarded and message.forwarded_from:
                    return

                
                if not allow_files and message.file:
                    return
                if not allow_stickers and message.sticker:
                    return
                if not allow_polls and message.poll:
                    return
                if not allow_contacts and message.contact_message:
                    return
                if not allow_locations and (message.location or message.live_location):
                    return

                
                if message.text:
                    text = message.text if case_sensitive else message.text.lower()
                    if min_text_length and len(message.text) < min_text_length:
                        return
                    if max_text_length and len(message.text) > max_text_length:
                        return
                    if contains and (contains if case_sensitive else contains.lower()) not in text:
                        return
                    if startswith and not text.startswith(startswith if case_sensitive else startswith.lower()):
                        return
                    if endswith and not text.endswith(endswith if case_sensitive else endswith.lower()):
                        return

                
                if commands:
                    if not message.text:
                        return
                    parts = message.text.strip().split()
                    cmd = parts[0].lstrip("/")
                    if cmd not in commands:
                        return
                    message.args = parts[1:]

                
                if filters and not filters(message):
                    return

                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "commands": commands,
                "chat_id": chat_id,
                "private_only": True,
                "sender_id": sender_id,
                "sender_type": sender_type
            })
            return wrapper

        return decorator
    def on_message_channel(
        self,
        chat_id: Optional[Union[str, List[str]]] = None,
        commands: Optional[List[str]] = None,
        filters: Optional[Callable[[Message], bool]] = None,
        sender_id: Optional[Union[str, List[str]]] = None,
        sender_type: Optional[str] = None,  
        allow_forwarded: bool = True,
        allow_files: bool = True,
        allow_stickers: bool = True,
        allow_polls: bool = True,
        allow_contacts: bool = True,
        allow_locations: bool = True,
        min_text_length: Optional[int] = None,
        max_text_length: Optional[int] = None,
        contains: Optional[str] = None,
        startswith: Optional[str] = None,
        endswith: Optional[str] = None,
        case_sensitive: bool = False
    ):
        """
        Sync decorator for handling only private messages with extended filters.
        """

        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                
                if not message.is_channel:
                    return

                
                if chat_id:
                    if isinstance(chat_id, str) and message.chat_id != chat_id:
                        return
                    if isinstance(chat_id, list) and message.chat_id not in chat_id:
                        return

                
                if sender_id:
                    if isinstance(sender_id, str) and message.sender_id != sender_id:
                        return
                    if isinstance(sender_id, list) and message.sender_id not in sender_id:
                        return

                
                if sender_type and message.sender_type != sender_type:
                    return

                
                if not allow_forwarded and message.forwarded_from:
                    return

                
                if not allow_files and message.file:
                    return
                if not allow_stickers and message.sticker:
                    return
                if not allow_polls and message.poll:
                    return
                if not allow_contacts and message.contact_message:
                    return
                if not allow_locations and (message.location or message.live_location):
                    return

                
                if message.text:
                    text = message.text if case_sensitive else message.text.lower()
                    if min_text_length and len(message.text) < min_text_length:
                        return
                    if max_text_length and len(message.text) > max_text_length:
                        return
                    if contains and (contains if case_sensitive else contains.lower()) not in text:
                        return
                    if startswith and not text.startswith(startswith if case_sensitive else startswith.lower()):
                        return
                    if endswith and not text.endswith(endswith if case_sensitive else endswith.lower()):
                        return

                
                if commands:
                    if not message.text:
                        return
                    parts = message.text.strip().split()
                    cmd = parts[0].lstrip("/")
                    if cmd not in commands:
                        return
                    message.args = parts[1:]

                
                if filters and not filters(message):
                    return

                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "commands": commands,
                "chat_id": chat_id,
                "private_only": True,
                "sender_id": sender_id,
                "sender_type": sender_type
            })
            return wrapper

        return decorator

    def on_message_group(
        self,
        chat_id: Optional[Union[str, List[str]]] = None,
        commands: Optional[List[str]] = None,
        filters: Optional[Callable[[Message], bool]] = None,
        sender_id: Optional[Union[str, List[str]]] = None,
        sender_type: Optional[str] = None,  
        allow_forwarded: bool = True,
        allow_files: bool = True,
        allow_stickers: bool = True,
        allow_polls: bool = True,
        allow_contacts: bool = True,
        allow_locations: bool = True,
        min_text_length: Optional[int] = None,
        max_text_length: Optional[int] = None,
        contains: Optional[str] = None,
        startswith: Optional[str] = None,
        endswith: Optional[str] = None,
        case_sensitive: bool = False
    ):
        """
        Sync decorator for handling only group messages with extended filters.
        """

        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                
                if not message.is_group:
                    return

                
                if chat_id:
                    if isinstance(chat_id, str) and message.chat_id != chat_id:
                        return
                    if isinstance(chat_id, list) and message.chat_id not in chat_id:
                        return

                
                if sender_id:
                    if isinstance(sender_id, str) and message.sender_id != sender_id:
                        return
                    if isinstance(sender_id, list) and message.sender_id not in sender_id:
                        return

                
                if sender_type and message.sender_type != sender_type:
                    return

                
                if not allow_forwarded and message.forwarded_from:
                    return

                
                if not allow_files and message.file:
                    return
                if not allow_stickers and message.sticker:
                    return
                if not allow_polls and message.poll:
                    return
                if not allow_contacts and message.contact_message:
                    return
                if not allow_locations and (message.location or message.live_location):
                    return

                
                if message.text:
                    text = message.text if case_sensitive else message.text.lower()
                    if min_text_length and len(message.text) < min_text_length:
                        return
                    if max_text_length and len(message.text) > max_text_length:
                        return
                    if contains and (contains if case_sensitive else contains.lower()) not in text:
                        return
                    if startswith and not text.startswith(startswith if case_sensitive else startswith.lower()):
                        return
                    if endswith and not text.endswith(endswith if case_sensitive else endswith.lower()):
                        return

                
                if commands:
                    if not message.text:
                        return
                    parts = message.text.strip().split()
                    cmd = parts[0].lstrip("/")
                    if cmd not in commands:
                        return
                    message.args = parts[1:]

                
                if filters and not filters(message):
                    return

                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "commands": commands,
                "chat_id": chat_id,
                "group_only": True,
                "sender_id": sender_id,
                "sender_type": sender_type
            })
            return wrapper

        return decorator

    def on_message(
        self, 
        filters: Optional[Callable[[Message], bool]] = None, 
        commands: Optional[List[str]] = None
    ):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if filters and not filters(message):
                    return
                if commands:
                    if not getattr(message, "is_command", False):
                        return
                    cmd = message.text.split()[0].lstrip("/") if message.text else ""
                    if cmd not in commands:
                        return
                return func(bot, message)
            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "commands": commands
            })
            return wrapper
        return decorator
    def on_message_file(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.file:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "file_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_forwarded(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.is_forwarded:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "forwarded_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_reply(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.is_reply:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "reply_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_text(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.text:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "text_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_media(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.is_media:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "media_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_sticker(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.sticker:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "sticker_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_contact(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.is_contact:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "contact_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_location(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.is_location:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "location_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def on_message_poll(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            def wrapper(bot, message: Message):
                if not message.is_poll:
                    return
                if filters and not filters(message):
                    return
                return func(bot, message)

            self._message_handlers.append({
                "func": wrapper,
                "filters": filters,
                "poll_only": True,
                "commands":commands
            })
            return wrapper
        return decorator

    def message_handler(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            self._message_handlers.append({
                "func": func,
                "filters": filters,
                "commands": commands
            })
            return func
        return decorator
    def on_update(self, filters: Optional[Callable[[Message], bool]] = None, commands: Optional[List[str]] = None):
        def decorator(func: Callable[[Any, Message], None]):
            self._message_handlers.append({
                "func": func,
                "filters": filters,
                "commands": commands
            })
            return func
        return decorator

    def on_callback(self, button_id: Optional[str] = None):
        def decorator(func: Callable[[Any, Message], None]):
            if not hasattr(self, "_callback_handlers"):
                self._callback_handlers = []
            self._callback_handlers.append({
                "func": func,
                "button_id": button_id
            })
            return func
        return decorator 
    def callback_query(self, button_id: Optional[str] = None):
        def decorator(func: Callable[[Any, Message], None]):
            if not hasattr(self, "_callback_handlers"):
                self._callback_handlers = []
            self._callback_handlers.append({
                "func": func,
                "button_id": button_id
            })
            return func
        return decorator
    def callback_query_handler(self, button_id: Optional[str] = None):
        def decorator(func: Callable[[Any, Message], None]):
            if not hasattr(self, "_callback_handlers"):
                self._callback_handlers = []
            self._callback_handlers.append({
                "func": func,
                "button_id": button_id
            })
            return func
        return decorator
    def _handle_inline_query(self, inline_message: InlineMessage):
        aux_button_id = inline_message.aux_data.button_id if inline_message.aux_data else None

        for handler in self._inline_query_handlers:
            if handler["button_id"] is None or handler["button_id"] == aux_button_id:
                try:
                    handler["func"](self, inline_message)
                except Exception as e:
                    print(f"Error in inline query handler: {e}")

    def on_inline_query(self, button_id: Optional[str] = None):
        def decorator(func: Callable[[Any, InlineMessage], None]):
            self._inline_query_handlers.append({
                "func": func,
                "button_id": button_id
            })
            return func
        return decorator

    
    def _process_update(self, update: dict):
        import threading

        if update.get("type") == "ReceiveQuery":
            msg = update.get("inline_message", {})
            context = InlineMessage(bot=self, raw_data=msg)

            
            if hasattr(self, "_callback_handlers"):
                for handler in self._callback_handlers:
                    cb_id = getattr(context.aux_data, "button_id", None)
                    if not handler["button_id"] or handler["button_id"] == cb_id:
                        threading.Thread(target=handler["func"], args=(self, context), daemon=True).start()

            
            threading.Thread(target=self._handle_inline_query, args=(context,), daemon=True).start()
            return

        if update.get("type") == "NewMessage":
            msg = update.get("new_message", {})
            try:
                if msg.get("time") and (time.time() - float(msg["time"])) > 20:
                    return
            except Exception:
                return

            context = Message(bot=self, 
                              chat_id=update.get("chat_id"), 
                              message_id=msg.get("message_id"), 
                              sender_id=msg.get("sender_id"), 
                              text=msg.get("text"), 
                              raw_data=msg)

            if context.aux_data and self._callback_handlers:
                for handler in self._callback_handlers:
                    if not handler["button_id"] or context.aux_data.button_id == handler["button_id"]:
                        threading.Thread(target=handler["func"], args=(self, context), daemon=True).start()
                        return

            if self._message_handlers:
                for handler in self._message_handlers:
                    if handler["commands"]:
                        if not context.text or not context.text.startswith("/"):
                            continue
                        parts = context.text.split()
                        cmd = parts[0][1:]
                        if cmd not in handler["commands"]:
                            continue
                        context.args = parts[1:]

                    if handler["filters"] and not handler["filters"](context):
                        continue

                    threading.Thread(target=handler["func"], args=(self, context), daemon=True).start()
                    continue

    def get_updates(
            self,
            offset_id: Optional[str] = None,
            limit: Optional[int] = None
    ) -> Dict[str, Any]:
        """
        Get updates from the Rubika Bot API.

        Parameters:
            offset_id (str, optional): The ID of the update to start receiving updates from. Defaults to None.
            limit (int, optional): The maximum number of updates to retrieve. Defaults to None.

        Returns:
            dict: The response containing the updates.
        """
        data = {}
        if offset_id:
            data["offset_id"] = offset_id
        if limit:
            data["limit"] = limit

        # Use the POST method to send the request and return the response
        return self._post("getUpdates", data)
    def update_webhook(
        self,
        offset_id: Optional[str] = None,
        limit: Optional[int] = None
    ) -> Dict[str, Any]:
        data = {}
        if offset_id:
            data["offset_id"] = offset_id
        if limit:
            data["limit"] = limit
        return list(requests.get(self.web_hook).json())
    def _is_duplicate(self, message_id: str, max_age_sec: int = 300) -> bool:
        now = time.time()

        expired = [mid for mid, ts in self._processed_message_ids.items() if now - ts > max_age_sec]
        for mid in expired:
            del self._processed_message_ids[mid]

        if message_id in self._processed_message_ids:
            return True

        self._processed_message_ids[message_id] = now
        return False

    def run(
        self,
        debug=False,
        sleep_time=0.1,
        webhook_timeout=20,
        update_limit=100,
        retry_delay=5,
        stop_on_error=False,
        max_errors=None,
        max_runtime=None,
        allowed_update_types=None,
        ignore_duplicate_messages=True,
        skip_inline_queries=False,
        skip_channel_posts=False,
        skip_service_messages=False,
        skip_edited_messages=False,
        skip_bot_messages=False,
        log_file=None,
        print_exceptions=True,
        error_handler=None,
        shutdown_hook=None,
        log_to_console=True,
        custom_update_fetcher=None,
        custom_update_processor=None,
        message_filter=None,
        notify_on_error=False,
        notification_handler=None,
    ):
        import time
        from typing import Dict
        if debug:
            print("[DEBUG] Bot started running server...")

        self._processed_message_ids: Dict[str, float] = {}
        error_count = 0
        start_time = time.time()

        try:
            while True:
                try:
                    
                    if max_runtime and (time.time() - start_time > max_runtime):
                        if debug:
                            print("[DEBUG] Max runtime reached, stopping...")
                        break

                    
                    if self.web_hook:
                        updates = custom_update_fetcher() if custom_update_fetcher else self.update_webhook()
                        if isinstance(updates, list):
                            for item in updates:
                                data = item.get("data", {})
                                received_at_str = item.get("received_at")

                                if received_at_str:
                                    try:
                                        received_at_ts = datetime.datetime.strptime(received_at_str, "%Y-%m-%d %H:%M:%S").timestamp()
                                        if time.time() - received_at_ts > webhook_timeout:
                                            continue
                                    except (ValueError, TypeError):
                                        pass

                                update = data.get("update") or (
                                    {"type": "ReceiveQuery", "inline_message": data.get("inline_message")}
                                    if "inline_message" in data else None
                                )
                                if not update:
                                    continue

                                
                                if skip_inline_queries and update.get("type") == "ReceiveQuery":
                                    continue
                                if skip_channel_posts and update.get("type") == "ChannelPost":
                                    continue
                                if skip_service_messages and update.get("type") == "ServiceMessage":
                                    continue
                                if skip_edited_messages and update.get("type") == "EditedMessage":
                                    continue
                                if skip_bot_messages and update.get("from", {}).get("is_bot"):
                                    continue
                                if allowed_update_types and update.get("type") not in allowed_update_types:
                                    continue

                                message_id = (
                                    update.get("new_message", {}).get("message_id")
                                    if update.get("type") == "NewMessage"
                                    else update.get("inline_message", {}).get("message_id")
                                    if update.get("type") == "ReceiveQuery"
                                    else update.get("message_id")
                                )

                                if message_id is not None:
                                    message_id = str(message_id)

                                if message_id and (not ignore_duplicate_messages or not self._is_duplicate(received_at_str)):
                                    if message_filter and not message_filter(update):
                                        continue
                                    if custom_update_processor:
                                        custom_update_processor(update)
                                    else:
                                        self._process_update(update)
                                    if message_id:
                                        self._processed_message_ids[message_id] = time.time()

                    
                    else:
                        updates = custom_update_fetcher() if custom_update_fetcher else self.get_updates(offset_id=self._offset_id, limit=update_limit)
                        if updates and updates.get("data"):
                            for update in updates["data"].get("updates", []):
                                if allowed_update_types and update.get("type") not in allowed_update_types:
                                    continue

                                message_id = (
                                    update.get("new_message", {}).get("message_id")
                                    if update.get("type") == "NewMessage"
                                    else update.get("inline_message", {}).get("message_id")
                                    if update.get("type") == "ReceiveQuery"
                                    else update.get("message_id")
                                )

                                if message_id is not None:
                                    message_id = str(message_id)

                                if message_id and (not ignore_duplicate_messages or not self._is_duplicate(message_id)):
                                    if message_filter and not message_filter(update):
                                        continue
                                    if custom_update_processor:
                                        custom_update_processor(update)
                                    else:
                                        self._process_update(update)
                                    if message_id:
                                        self._processed_message_ids[message_id] = time.time()

                            self._offset_id = updates["data"].get("next_offset_id", self._offset_id)

                    if sleep_time:
                        time.sleep(sleep_time)

                except Exception as e:
                    error_count += 1
                    if log_to_console:
                        print(f"Error in run loop: {e}")
                    if log_file:
                        with open(log_file, "a", encoding="utf-8") as f:
                            f.write(f"{datetime.datetime.now()} - ERROR: {e}\n")
                    if print_exceptions:
                        import traceback
                        traceback.print_exc()
                    if error_handler:
                        error_handler(e)
                    if notify_on_error and notification_handler:
                        notification_handler(e)

                    if max_errors and error_count >= max_errors and stop_on_error:
                        break

                    time.sleep(retry_delay)

        finally:
            if shutdown_hook:
                shutdown_hook()
            if debug:
                print("Bot stopped and session closed.")

    def send_message(
        self,
        chat_id: str,
        text: str,
        chat_keypad: Optional[Dict[str, Any]] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        disable_notification: bool = False,
        reply_to_message_id: Optional[str] = None,
        chat_keypad_type: Optional[Literal["New", "Removed"]] = None,
        delete_after = None,
        parse_mode = None
    ) -> Dict[str, Any]:
        """
        Send a text message to a chat.
        """
        payload = {
            "chat_id": chat_id,
            "text": text,
            "disable_notification": disable_notification
        }
        if chat_keypad:
            payload["chat_keypad"] = chat_keypad
        if inline_keypad:
            payload["inline_keypad"] = inline_keypad
        if reply_to_message_id:
            payload["reply_to_message_id"] = reply_to_message_id
        if chat_keypad_type:
            payload["chat_keypad_type"] = chat_keypad_type

        return self._post("sendMessage", payload)

    def _get_client(self):
        if self.session_name:
            return Client_get(self.session_name,self.auth,self.Key,self.platform)
        else :
            return Client_get(show_last_six_words(self.token),self.auth,self.Key,self.platform)
    from typing import Union

    def check_join(self, channel_guid: str, chat_id: str = None) -> Union[bool, list[str]]:
        client = self._get_client()

        if chat_id:
            chat_info = self.get_chat(chat_id).get('data', {}).get('chat', {})
            username = chat_info.get('username')
            user_id = chat_info.get('user_id')

            if username:
                members = self.get_all_member(channel_guid, search_text=username).get('in_chat_members', [])
                return any(m.get('username') == username for m in members)

            elif user_id:
                member_guids = client.get_all_members(channel_guid, just_get_guids=True)
                return user_id in member_guids

            return False

        return False

    def get_url_file(self,file_id):
        data = self._post("getFile", {'file_id': file_id})
        return data.get("data").get("download_url")
    

    def get_all_member(
        self,
        channel_guid: str,
        search_text: str = None,
        start_id: str = None,
        just_get_guids: bool = False
    ):
        client = self._get_client()
        return client.get_all_members(channel_guid, search_text, start_id, just_get_guids)

    def send_poll(
        self,
        chat_id: str,
        question: str,
        options: List[str]
    ) -> Dict[str, Any]:
        """
        Send a poll to a chat.
        """
        return self._post("sendPoll", {
            "chat_id": chat_id,
            "question": question,
            "options": options
        })

    def send_location(
            self,
            chat_id: str,
            latitude: float,
            longitude: float,
            disable_notification: bool = False,
            inline_keypad: Optional[Dict[str, Any]] = None,
            reply_to_message_id: Optional[str] = None,
            chat_keypad_type: Optional[Literal["New", "Removed"]] = None
    ) -> Dict[str, Any]:
        """
        Send a location to a chat.

        Parameters:
            chat_id (str): The unique identifier for the target chat.
            latitude (float): The latitude of the location.
            longitude (float): The longitude of the location.
            disable_notification (bool, optional): If True, the message will be sent without a notification. Defaults to False.
            inline_keypad (dict, optional): Inline keypad options. Defaults to None.
            reply_to_message_id (str, optional): If the message is a reply, this parameter is the ID of the original message. Defaults to None.
            chat_keypad_type (Literal["New", "Removed"], optional): The type of keypad (New or Removed). Defaults to None.

        Returns:
            dict: The response from the API after sending the location.
        """
        payload = {
            "chat_id": chat_id,
            "latitude": latitude,
            "longitude": longitude,
            "disable_notification": disable_notification,
            "inline_keypad": inline_keypad,
            "reply_to_message_id": reply_to_message_id,
            "chat_keypad_type": chat_keypad_type
        }
        payload = {k: v for k, v in payload.items() if v is not None}
        return self._post("sendLocation", payload)

    def send_contact(
        self,
        chat_id: str,
        first_name: str,
        last_name: str,
        phone_number: str
    ) -> Dict[str, Any]:
        """
        Send a contact to a chat.
        """
        return self._post("sendContact", {
            "chat_id": chat_id,
            "first_name": first_name,
            "last_name": last_name,
            "phone_number": phone_number
        })
    def download(self,file_id: str, save_as: str = None, chunk_size: int = 1024 * 512, timeout_sec: int = 60, verbose: bool = False):
        """
        Download a file from server using its file_id with chunked transfer,
        progress bar, file extension detection, custom filename, and timeout.

        If save_as is not provided, filename will be extracted from
        Content-Disposition header or Content-Type header extension.

        Parameters:
            file_id (str): The file ID to fetch the download URL.
            save_as (str, optional): Custom filename to save. If None, automatically detected.
            chunk_size (int, optional): Size of each chunk in bytes. Default 512KB.
            timeout_sec (int, optional): HTTP timeout in seconds. Default 60.
            verbose (bool, optional): Show progress messages. Default True.

        Returns:
            bool: True if success, raises exceptions otherwise.
        """
        try:
            url = self.get_url_file(file_id)
            if not url:
                raise ValueError("Download URL not found in response.")
        except Exception as e:
            raise ValueError(f"Failed to get download URL: {e}")

        try:
            with requests.get(url, stream=True, timeout=timeout_sec) as resp:
                if resp.status_code != 200:
                    raise requests.HTTPError(f"Failed to download file. Status code: {resp.status_code}")

                if not save_as:
                    content_disp = resp.headers.get("Content-Disposition", "")
                    match = re.search(r'filename="?([^\";]+)"?', content_disp)
                    if match:
                        save_as = match.group(1)
                    else:
                        content_type = resp.headers.get("Content-Type", "").split(";")[0]
                        extension = mimetypes.guess_extension(content_type) or ".bin"
                        save_as = f"{file_id}{extension}"

                total_size = int(resp.headers.get("Content-Length", 0))
                progress = tqdm(total=total_size, unit="B", unit_scale=True)

                with open(save_as, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=chunk_size):
                        if chunk:
                            f.write(chunk)
                            progress.update(len(chunk))

                progress.close()
                if verbose:
                    print(f"File saved as: {save_as}")

                return True

        except Exception as e:
            raise RuntimeError(f"Download failed: {e}")
    def get_chat(self, chat_id: str) -> Dict[str, Any]:
        """Get chat info."""
        return self._post("getChat", {"chat_id": chat_id})

    def upload_media_file(self, upload_url: str, name: str, path: Union[str, Path]) -> str:
        is_temp_file = False

        if isinstance(path, str) and path.startswith("http"):
            response = requests.get(path)
            if response.status_code != 200:
                raise Exception(f"Failed to download file from URL ({response.status_code})")
            temp_file = tempfile.NamedTemporaryFile(delete=False)
            temp_file.write(response.content)
            temp_file.close()
            path = temp_file.name
            is_temp_file = True

        file_size = os.path.getsize(path)

        with open(path, 'rb') as f:
            progress_bar = None

            if self.show_progress:
                progress_bar = tqdm(
                    total=file_size,
                    unit='B',
                    unit_scale=True,
                    unit_divisor=1024,
                    desc=f'Uploading : {name}',
                    bar_format='{l_bar}{bar:100}{r_bar}',
                    colour='cyan'
                )

            class FileWithProgress:
                def __init__(self, file, progress):
                    self.file = file
                    self.progress = progress

                def read(self, size=-1):
                    data = self.file.read(size)
                    if self.progress:
                        self.progress.update(len(data))
                    return data

                def __getattr__(self, attr):
                    return getattr(self.file, attr)

            file_with_progress = FileWithProgress(f, progress_bar)

            files = {
                'file': (name, file_with_progress, 'application/octet-stream')
            }

            response = requests.post(upload_url, files=files)

            if progress_bar:
                progress_bar.close()

        if is_temp_file:
            os.remove(path)

        if response.status_code != 200:
            raise Exception(f"Upload failed ({response.status_code}): {response.text}")

        data = response.json()
        return data.get('data', {}).get('file_id')

    def send_button_join(
    self, 
    chat_id, 
    title_button :  Union[str, list], 
    username :  Union[str, list], 
    text, 
    reply_to_message_id=None, 
    id="None"
):
        from .button import InlineBuilder
        builder = InlineBuilder()

        
        if isinstance(username, (list, tuple)) and isinstance(title_button, (list, tuple)):
            for t, u in zip(title_button, username):
                builder = builder.row(
                    InlineBuilder().button_join_channel(
                        text=t,
                        id=id,
                        username=u
                    )
                )

        
        elif isinstance(username, (list, tuple)) and isinstance(title_button, str):
            for u in username:
                builder = builder.row(
                    InlineBuilder().button_join_channel(
                        text=title_button,
                        id=id,
                        username=u
                    )
                )

        
        else:
            builder = builder.row(
                InlineBuilder().button_join_channel(
                    text=title_button,
                    id=id,
                    username=username
                )
            )

        return self.send_message(
            chat_id=chat_id,
            text=text,
            inline_keypad=builder.build(),
            reply_to_message_id=reply_to_message_id
        )


    def send_button_url(
        self, 
        chat_id, 
        title_button :  Union[str, list], 
        url :  Union[str, list], 
        text, 
        reply_to_message_id=None, 
        id="None"
    ):
        from .button import InlineBuilder
        builder = InlineBuilder()

        
        if isinstance(url, (list, tuple)) and isinstance(title_button, (list, tuple)):
            for t, u in zip(title_button, url):
                builder = builder.row(
                    InlineBuilder().button_url_link(
                        text=t,
                        id=id,
                        url=u
                    )
                )

        
        elif isinstance(url, (list, tuple)) and isinstance(title_button, str):
            for u in url:
                builder = builder.row(
                    InlineBuilder().button_url_link(
                        text=title_button,
                        id=id,
                        url=u
                    )
                )

        
        else:
            builder = builder.row(
                InlineBuilder().button_url_link(
                    text=title_button,
                    id=id,
                    url=url
                )
            )

        return self.send_message(
            chat_id=chat_id,
            text=text,
            inline_keypad=builder.build(),
            reply_to_message_id=reply_to_message_id
        )


    def get_upload_url(self, media_type: Literal['File', 'Image', 'Voice', 'Music', 'Gif','Video']) -> str:
        allowed = ['File', 'Image', 'Voice', 'Music', 'Gif','Video']
        if media_type not in allowed:
            raise ValueError(f"Invalid media type. Must be one of {allowed}")
        result = self._post("requestSendFile", {"type": media_type})
        return result.get("data", {}).get("upload_url")
    def _send_uploaded_file(self, chat_id: str, file_id: str,type_file : str = "file",text: Optional[str] = None, chat_keypad: Optional[Dict[str, Any]] = None, inline_keypad: Optional[Dict[str, Any]] = None, disable_notification: bool = False, reply_to_message_id: Optional[str] = None, chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None") -> Dict[str, Any]:
        payload = {
            "chat_id": chat_id,
            "file_id": file_id,
            "text": text,
            "disable_notification": disable_notification,
            "chat_keypad_type": chat_keypad_type,
        }
        if chat_keypad:
            payload["chat_keypad"] = chat_keypad
        if inline_keypad:
            payload["inline_keypad"] = inline_keypad
        if reply_to_message_id:
            payload["reply_to_message_id"] = str(reply_to_message_id)

        resp = self._post("sendFile", payload)
        message_id_put = resp["data"]["message_id"]
        result = {
            "status": resp.get("status"),
            "status_det": resp.get("status_det"),
            "file_id": file_id,
            "text":text,
            "message_id": message_id_put,
            "send_to_chat_id": chat_id,
            "reply_to_message_id": reply_to_message_id,
            "disable_notification": disable_notification,
            "type_file": type_file,
            "raw_response": resp,
            "chat_keypad":chat_keypad,
            "inline_keypad":inline_keypad,
            "chat_keypad_type":chat_keypad_type
        }
        import json
        return json.dumps(result, ensure_ascii=False, indent=4)
    def send_file(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("File")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=caption,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def re_send_file(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        caption: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("File")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=caption,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def send_document(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        text: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("File")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def send_music(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        text: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("Music")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def send_video(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        text: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("Video")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def send_voice(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        text: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("Voice")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def send_image(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        text: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("Image")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )
    def send_gif(
        self,
        chat_id: str,
        path: Optional[Union[str, Path]] = None,
        file_id: Optional[str] = None,
        text: Optional[str] = None,
        file_name: Optional[str] = None,
        inline_keypad: Optional[Dict[str, Any]] = None,
        chat_keypad: Optional[Dict[str, Any]] = None,
        reply_to_message_id: Optional[str] = None,
        disable_notification: bool = False,
        chat_keypad_type: Optional[Literal["New", "Removed", "None"]] = "None"
    ) -> Dict[str, Any]:
        if path:
            file_name = file_name or Path(path).name
            upload_url = self.get_upload_url("Gif")
            file_id = self.upload_media_file(upload_url, file_name, path)
        if not file_id:
            raise ValueError("Either path or file_id must be provided.")
        return self._send_uploaded_file(
            chat_id=chat_id,
            file_id=file_id,
            text=text,
            inline_keypad=inline_keypad,
            chat_keypad=chat_keypad,
            reply_to_message_id=reply_to_message_id,
            disable_notification=disable_notification,
            chat_keypad_type=chat_keypad_type
        )

    

    def forward_message(
        self,
        from_chat_id: str,
        message_id: str,
        to_chat_id: str,
        disable_notification: bool = False
    ) -> Dict[str, Any]:
        """Forward a message from one chat to another."""
        return self._post("forwardMessage", {
            "from_chat_id": from_chat_id,
            "message_id": message_id,
            "to_chat_id": to_chat_id,
            "disable_notification": disable_notification
        })

    def edit_message_text(
        self,
        chat_id: str,
        message_id: str,
        text: str
    ) -> Dict[str, Any]:
        """Edit text of an existing message."""
        return self._post("editMessageText", {
            "chat_id": chat_id,
            "message_id": message_id,
            "text": text
        })

    def edit_inline_keypad(
        self,
        chat_id: str,
        message_id: str,
        inline_keypad: Dict[str, Any],
        text : str = None
    ) -> Dict[str, Any]:
        """Edit inline keypad of a message."""
        if text is not None:self._post("editMessageText", {"chat_id": chat_id,"message_id": message_id,"text": text})
        return self._post("editMessageKeypad", {
            "chat_id": chat_id,
            "message_id": message_id,
            "inline_keypad": inline_keypad
        })

    def delete_message(self, chat_id: str, message_id: str) -> Dict[str, Any]:
        """Delete a message from chat."""
        return self._post("deleteMessage", {
            "chat_id": chat_id,
            "message_id": message_id
        })

    def set_commands(self, bot_commands: List[Dict[str, str]]) -> Dict[str, Any]:
        """Set bot commands."""
        return self._post("setCommands", {"bot_commands": bot_commands})

    def update_bot_endpoint(self, url: str, type: str) -> Dict[str, Any]:
        """Update bot endpoint (Webhook or Polling)."""
        return self._post("updateBotEndpoints", {
            "url": url,
            "type": type
        })

    def remove_keypad(self, chat_id: str) -> Dict[str, Any]:
        """Remove chat keypad."""
        return self._post("editChatKeypad", {
            "chat_id": chat_id,
            "chat_keypad_type": "Removed"
        })

    def edit_chat_keypad(self, chat_id: str, chat_keypad: Dict[str, Any]) -> Dict[str, Any]:
        """Edit or add new chat keypad."""
        return self._post("editChatKeypad", {
            "chat_id": chat_id,
            "chat_keypad_type": "New",
            "chat_keypad": chat_keypad
        })
    def get_name(self, chat_id: str) -> str:
        try:
            chat = self.get_chat(chat_id)
            chat_info = chat.get("data", {}).get("chat", {})
            first_name = chat_info.get("first_name", "")
            last_name = chat_info.get("last_name", "")
            
            if first_name and last_name:
                return f"{first_name} {last_name}"
            elif first_name:
                return first_name
            elif last_name:
                return last_name
            else:
                return "Unknown"
        except Exception:
            return "Unknown"
    def get_username(self, chat_id: str) -> str:
        chat_info = self.get_chat(chat_id).get("data", {}).get("chat", {})
        return chat_info.get("username", "None")


import asyncio
import random
from typing import Optional, Dict, Any

import httpx

class TV:
    BASE_URL_TEMPLATE = "https://rbvod{}.iranlms.ir/"
    META_URL = "https://tv.rubika.ir/meta.json"
    
    DEFAULT_HEADERS = {
        'accept': '*/*',
        'accept-encoding': 'gzip, deflate, br, zstd',
        'accept-language': 'fa',
        'content-type': 'application/json;charset=UTF-8',
        'origin': 'https://tv.rubika.ir',
        'referer': 'https://tv.rubika.ir/',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/128.0.0.0 Safari/537.36'
    }

    def __init__(
        self,
        auth: str,
        proxy: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None
    ) -> None:
        """
        - auth: توکن احراز هویت حساب کاربری
        - proxy: پراکسی برای ارسال درخواست‌ها (اختیاری)
        - headers: هدرهای سفارشی برای ارسال درخواست‌ها (اختیاری)
        # نکته:
        https://vod.rubika.ir/ 
        """
        self.auth = auth
        self.headers = headers if headers is not None else self.DEFAULT_HEADERS
        self._version: Optional[str] = None
        self.client = httpx.AsyncClient(
            headers=self.headers,
            proxies=proxy,
            timeout=15.0,
            http2=True
        )

    async def _get_version(self) -> str:
        if self._version is None:
            try:
                response = await self.client.get(self.META_URL)
                response.raise_for_status()
                self._version = response.json()['version']
            except httpx.RequestError:
                self._version = "3.2.3"
        return self._version

    def _get_request_url(self) -> str:
        return self.BASE_URL_TEMPLATE.format(random.randint(1, 4))

    async def _request_post(self, method: str, data: Dict[str, Any]) -> Dict[str, Any]:
        payload = {
            "auth": self.auth,
            "api_version": "1",
            "client": {
                "app_name": "Main",
                "app_version": await self._get_version(),
                "package": "tv.rubika.ir",
                "platform": "TVWeb",
                "lang_code": "fa"
            },
            "data": data,
            "method": method
        }
        
        try:
            response = await self.client.post(self._get_request_url(), json=payload)
            response.raise_for_status()  
            response_data = response.json()
            
            if 'data' in response_data:
                return response_data['data']
            elif 'status_det' in response_data:
                return {'status_detail': response_data['status_det']}
            raise ValueError("پاسخ نامعتبر از سرور دریافت شد.")
        
        except httpx.RequestError as e:
            
            raise ConnectionError(f"خطا در ارتباط با سرور: {e}") from e

    
    

    async def get_list_items(self, list_id: str, type: str = "Media", start_id: str = "0"):
        return await self._request_post("getListItems", {"list_id": list_id, "type": type, "start_id": start_id})

    async def get_media_by_id(self, media_id: str,track_id:str=None):
        return await self._request_post("getMedia", {"media_id": media_id,"track_id":track_id})

    async def get_custom_menu_items(self):
        return await self._request_post("getCustomMenuItems", {})
    
    async def get_wish_list(self, start_id: Optional[str] = None):
        return await self._request_post("getWishList", {"start_id": start_id})
    
    async def get_Property_Types(self):
        return await self._request_post("getPropertyTypes", {})
    
    async def get_Listing(self, listing_id: Optional[str] = "home"):
        return await self._request_post("getListing", {"listing_id": listing_id})
    
    async def get_me(self):
        return await self._request_post("getAccountInfo", {})

    async def search_video(self, text: str, start_id: Optional[str] = None):
        return await self._request_post("search", {'search_text': text, 'start_id': start_id})

    async def _action_on_media(self, action: str, media_id: str):
        
        return await self._request_post('actionOnMedia', {'action': action, 'media_id': media_id})

    async def like_media(self, media_id: str):
        return await self._action_on_media("Like", media_id)

    async def un_like_media(self, media_id: str):
        return await self._action_on_media("Dislike", media_id)

    async def add_wish_media(self, media_id: str):
        return await self._action_on_media("AddWishList", media_id)

    async def get_cast_medias(self, cast_id: str, start_id: str = "0"):
        return await self._request_post("getCastMedias", {'cast_id': cast_id, 'start_id': start_id})

    async def get_cast_info(self, cast_id: str):
        return await self._request_post("getCastDetails", {'cast_id': cast_id})
    async def get_Season_Episodes(self, season_id: str,series_id:str):
        return await self._request_post("getSeasonEpisodes", {'season_id': season_id,"series_id":series_id})
    async def get_Related(self, media_id: str,start_id:str="0"):
        return await self._request_post("getRelated", {'media_id': media_id,"start_id":start_id})

    
    async def close(self):
        await self.client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()